"""MIST-compatible MedNeXt model."""
from typing import List, Union, Optional
import torch.nn as nn
import torch.nn.functional as F

from mist.models.mednext_v1 import blocks


class MedNeXt(nn.Module):
    def __init__(self,
        in_channels: int,
        n_channels: int,
        n_classes: int,
        exp_r: Union[List[int], int]=4,
        kernel_size: int=7,
        enc_kernel_size: Optional[int]=None,
        dec_kernel_size: Optional[int]=None,
        deep_supervision: bool=False,
        pocket: bool=False,
        do_res: bool=False,
        do_res_up_down: bool=False,
        block_counts: list=[2,2,2,2,2,2,2,2,2],
        norm_type: str='group',
        dim: str='3d',
        grn: bool=False
    ):

        super().__init__()

        self.do_ds = deep_supervision
        assert dim in ['2d', '3d']

        if kernel_size is not None:
            enc_kernel_size = kernel_size
            dec_kernel_size = kernel_size

        if dim == '2d':
            conv = nn.Conv2d
        elif dim == '3d':
            conv = nn.Conv3d

        self.stem = conv(in_channels, n_channels, kernel_size=1)
        if isinstance(exp_r, int):
            exp_r = [exp_r] * len(block_counts)

        self.enc_block_0 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels,
                out_channels=n_channels,
                exp_r=exp_r[0],
                kernel_size=enc_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                ) 
            for i in range(block_counts[0])]
        ) 

        self.down_0 = blocks.MedNeXtDownBlock(
            in_channels=n_channels,
            out_channels=n_channels if pocket else 2*n_channels,
            exp_r=exp_r[1],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim
        )
    
        self.enc_block_1 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*2,
                out_channels=n_channels if pocket else n_channels*2,
                exp_r=exp_r[1],
                kernel_size=enc_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[1])]
        )

        self.down_1 = blocks.MedNeXtDownBlock(
            in_channels=n_channels if pocket else 2*n_channels,
            out_channels=n_channels if pocket else 4*n_channels,
            exp_r=exp_r[2],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )

        self.enc_block_2 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*4,
                out_channels=n_channels if pocket else n_channels*4,
                exp_r=exp_r[2],
                kernel_size=enc_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[2])]
        )

        self.down_2 = blocks.MedNeXtDownBlock(
            in_channels=n_channels if pocket else 4*n_channels,
            out_channels=n_channels if pocket else 8*n_channels,
            exp_r=exp_r[3],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )
        
        self.enc_block_3 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*8,
                out_channels=n_channels if pocket else n_channels*8,
                exp_r=exp_r[3],
                kernel_size=enc_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )            
            for i in range(block_counts[3])]
        )
        
        self.down_3 = blocks.MedNeXtDownBlock(
            in_channels=n_channels if pocket else 8*n_channels,
            out_channels=n_channels if pocket else 16*n_channels,
            exp_r=exp_r[4],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )

        self.bottleneck = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*16,
                out_channels=n_channels if pocket else n_channels*16,
                exp_r=exp_r[4],
                kernel_size=dec_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[4])]
        )

        self.up_3 = blocks.MedNeXtUpBlock(
            in_channels=n_channels if pocket else 16*n_channels,
            out_channels=n_channels if pocket else 8*n_channels,
            exp_r=exp_r[5],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )

        self.dec_block_3 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*8,
                out_channels=n_channels if pocket else n_channels*8,
                exp_r=exp_r[5],
                kernel_size=dec_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[5])]
        )

        self.up_2 = blocks.MedNeXtUpBlock(
            in_channels=n_channels if pocket else 8*n_channels,
            out_channels=n_channels if pocket else 4*n_channels,
            exp_r=exp_r[6],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )

        self.dec_block_2 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*4,
                out_channels=n_channels if pocket else n_channels*4,
                exp_r=exp_r[6],
                kernel_size=dec_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[6])]
        )

        self.up_1 = blocks.MedNeXtUpBlock(
            in_channels=n_channels if pocket else 4*n_channels,
            out_channels=n_channels if pocket else 2*n_channels,
            exp_r=exp_r[7],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )

        self.dec_block_1 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels if pocket else n_channels*2,
                out_channels=n_channels if pocket else n_channels*2,
                exp_r=exp_r[7],
                kernel_size=dec_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[7])]
        )

        self.up_0 = blocks.MedNeXtUpBlock(
            in_channels=n_channels if pocket else 2*n_channels,
            out_channels=n_channels,
            exp_r=exp_r[8],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn
        )

        self.dec_block_0 = nn.Sequential(*[
            blocks.MedNeXtBlock(
                in_channels=n_channels,
                out_channels=n_channels,
                exp_r=exp_r[8],
                kernel_size=dec_kernel_size,
                do_res=do_res,
                norm_type=norm_type,
                dim=dim,
                grn=grn
                )
            for i in range(block_counts[8])]
        )

        self.out_0 = blocks.OutBlock(in_channels=n_channels, n_classes=n_classes, dim=dim)

        if deep_supervision:
            self.out_1 = blocks.OutBlock(in_channels=n_channels if pocket else n_channels*2, n_classes=n_classes, dim=dim)
            self.out_2 = blocks.OutBlock(in_channels=n_channels if pocket else n_channels*4, n_classes=n_classes, dim=dim)
            self.out_3 = blocks.OutBlock(in_channels=n_channels if pocket else n_channels*8, n_classes=n_classes, dim=dim)
            self.out_4 = blocks.OutBlock(in_channels=n_channels if pocket else n_channels*16, n_classes=n_classes, dim=dim)

        self.block_counts = block_counts


    def forward(self, x):
        x = self.stem(x)
        x_res_0 = self.enc_block_0(x)
        x = self.down_0(x_res_0)
        x_res_1 = self.enc_block_1(x)
        x = self.down_1(x_res_1)
        x_res_2 = self.enc_block_2(x)
        x = self.down_2(x_res_2)
        x_res_3 = self.enc_block_3(x)
        x = self.down_3(x_res_3)

        x = self.bottleneck(x)
        if self.do_ds and self.training:
            x_ds_4 = self.out_4(x)

        x_up_3 = self.up_3(x)
        dec_x = x_res_3 + x_up_3
        x = self.dec_block_3(dec_x)

        if self.do_ds and self.training:
            x_ds_3 = self.out_3(x)
        del x_res_3, x_up_3

        x_up_2 = self.up_2(x)
        dec_x = x_res_2 + x_up_2
        x = self.dec_block_2(dec_x)
        if self.do_ds and self.training:
            x_ds_2 = self.out_2(x)
        del x_res_2, x_up_2

        x_up_1 = self.up_1(x)
        dec_x = x_res_1 + x_up_1
        x = self.dec_block_1(dec_x)
        if self.do_ds and self.training:
            x_ds_1 = self.out_1(x)
        del x_res_1, x_up_1

        x_up_0 = self.up_0(x)
        dec_x = x_res_0 + x_up_0
        x = self.dec_block_0(dec_x)
        del x_res_0, x_up_0, dec_x

        x = self.out_0(x)

        # Make MedNeXt compatible with MIST.
        if self.training:
            output = {}
            output["prediction"] = x

            if self.do_ds:
                patch_size = x.shape[2:]
                x_ds_1 = F.interpolate(x_ds_1, size=patch_size)
                x_ds_2 = F.interpolate(x_ds_2, size=patch_size)
                x_ds_3 = F.interpolate(x_ds_3, size=patch_size)
                x_ds_4 = F.interpolate(x_ds_4, size=patch_size)
                output["deep_supervision"] = (x_ds_1, x_ds_2, x_ds_3, x_ds_4)
        else:
            output = x

        return output
